#pragma once
#include <libssh/sftp.h>
#include <libssh/libsshpp.hpp>
#include "TCPSession.h"

typedef boost::asio::io_service::strand io_strand;

class SFTPChannel;
typedef std::shared_ptr<SFTPChannel> sftp_channel_ptr;

class SSHSession : public ssh::Session, public TCPSession
{
private:
	friend SFTPChannel;
	static int srecv(socket_t /*socket*/, void *userdata, char *buffer, int size);
	static int ssend(socket_t /*socket*/, void *userdata, const char *buffer, int size);
	static int sclose(socket_t /*socket*/, void *userdata);
	int recv(char *buffer, int size);
	int send(const char *buffer, int size);
	int closecb();
	std::unique_ptr<yield_context> yield_;
	SSHSession(io_service &service);

	void throw_null(void *ptr);

	bool throw_error(ssize_t error);

	void yielder(yield_context &yield);

public:

	io_strand strand_;

	void Logon(yield_context &yield,
		const char *host, unsigned short port,
		const char *username, const char *passwd);

	sftp_channel_ptr SSHSession::CreateSFTP(yield_context &yield);

	static std::shared_ptr<SSHSession> Create(io_service &service);

	std::future<void> Logon(const char *host, unsigned short port, const char *username, const char *passwd);

	std::future<sftp_channel_ptr> CreateSFTP();

	ssh_session GetSession();
};


class SFTPFile
{
private:
	sftp_channel_ptr channel_;
	//JQS this might not work. yield pointer has to be populated for sftp_close to work
	sftp_file file_;

public:
	SFTPFile(sftp_channel_ptr channel, sftp_file file);
	~SFTPFile();
};

class SFTPDirectory
{
private:
	sftp_channel_ptr channel_;
	sftp_dir  dir_;
public:
	SFTPDirectory(sftp_channel_ptr channel, sftp_dir dir);
	~SFTPDirectory();
};

typedef std::unique_ptr<SFTPFile> sftp_file_ptr;
typedef std::unique_ptr<SFTPDirectory> sftp_dir_ptr;

class SFTPChannel : public std::enable_shared_from_this<SFTPChannel>
{
private:
	friend class SSHSession;

	std::string last_message_;
	int last_status_;
	std::shared_ptr<SSHSession> session_;
	sftp_session sftp_;
	friend class SFTPFile;

	SFTPChannel(std::shared_ptr<SSHSession> session, sftp_session sftp);

	void throw_exception();
	void throw_null(void *ptr);
	bool throw_error(ssize_t error);

public:
	~SFTPChannel();

	sftp_file_ptr OpenFile(yield_context &_yield, const char *filename, int access, mode_t mode = 0666);

	sftp_dir_ptr OpenDir(yield_context &_yield, const char *path);

	std::future<sftp_file_ptr> OpenFile(const char *filename, int access, mode_t mode = 0666);
	std::future<sftp_dir_ptr> OpenDir(const char *path);
};
